/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.awt.event.*;
import javax.swing.*;
import java.lang.reflect.*;
//import javax.comm.*;
/**
 * CCG: Twidor- The Twiddler Tutor!
 * <pre>
 * EventHandler.java, the general event manager for Twidor.
 *
 * Revisions:
 * 			0.5	17 July 2003
 * 			Completed Tutor
 * 			0.1	22 May 2003
 * 			Created class EventHandler
 * </pre>
 * @author <a href="mailto:visyz@cc.gatech.edu">Gibby Fusia</a>
 * @version Version 0.5; 17 July 2003
 */
public class EventHandler implements ActionListener, ItemListener,
	KeyListener, WindowListener {

	/**
	 * internal variables
	 */
	private TwidorMain theTutor;
	private SerialInterface myTwiddler;
	private TwidorConstants myDefaults;
	private Timer myHighlightTimer;
	private Timer myMultiTapTimer;

	private boolean enabled = true;

	/**
	 * Default Constructor.
	 * Private. Don't want anybody to use it.
	 */
	private EventHandler () {
		theTutor = null;
		myTwiddler = null;
		myDefaults = null;
		myHighlightTimer = null;
		myMultiTapTimer = null;
	}// end EventHandler

	/**
	 * Default Constructor.
	 * @param tutor the tutor we manage events for
	 * @param twidorDefaults the defaults we adhere to
	 */
	public EventHandler (TwidorMain tutor, TwidorConstants twidorDefaults) {
		this();
		theTutor = tutor;
		myDefaults = twidorDefaults;
		init();
	}// end EventHandler

	/**
	 * Called when we want to change the internals to EventHandler
	 */
	public void init () {
		if (getDefaults().USE_SERIAL) {
			//myTwiddler = new SerialTwiddler(getDefaults(), this);

			//	Do late binding so that it can run on machines without javax.comm (ie macs)
			try {
				Class serialClass = Class.forName("SerialTwiddler");
				Class [] classList = new Class[] {Class.forName("TwidorConstants"), Class.forName("EventHandler")};
				//System.out.println(classList);
				Constructor serialConstructor = serialClass.getConstructor(classList);
				myTwiddler = (SerialInterface) serialConstructor.newInstance(new Object[] {getDefaults(), this});
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println();
				System.out.println("Error: Unable to initialize serial interface.");
				System.exit(1);
			}
			System.out.println("Serial interface initialized.");
		}
		initMultiTap();

		if (getDefaults().HIGHLIGHT_HINT && getDefaults().HIGHLIGHT_DELAY > 0) {
			myHighlightTimer = new Timer(getDefaults().HIGHLIGHT_DELAY, this);
			myHighlightTimer.setRepeats(false);
			myHighlightTimer.setCoalesce(false);
		}
	}// end init ()

	/**
	 * Called when we want to change if we're using multitap or not without reinitializing the serial port
	 */
	public void reinit() {
		if (getDefaults().USE_MULTITAP) {
			initMultiTap();
		} else {
			if (myMultiTapTimer != null) {
				myMultiTapTimer.stop();
				myMultiTapTimer = null;
			}
		}
	}

	public void initMultiTap() {
		if (getDefaults().USE_MULTITAP) {
			myMultiTapTimer = new Timer(getDefaults().MULTITAP_TIMEOUT, this);
			myMultiTapTimer.setRepeats(false);
			myMultiTapTimer.setCoalesce(false);
		}
	}

	/**
	 * invoked when the user wants to reset the EventHandler
	 */
	public void cleanup () {
		if (getDefaults().USE_SERIAL) {
			getTwiddler().close();
		}
		if (myMultiTapTimer != null) {
			myMultiTapTimer.stop();
			myMultiTapTimer = null;
		}
		if (myHighlightTimer != null) {
			myHighlightTimer.stop();
			myHighlightTimer = null;
		}
		enabled = false;
	}// end cleanup ()

	/**
	 * feedback loop to stop the multitap timer on 0,1,*,# keys
	 */
	public void multiStop () {
		if (getDefaults().bDEBUG)
			System.out.println("EventHandler: Stopping multitap timer");
		if (myMultiTapTimer != null) {
			myMultiTapTimer.stop();
		}
	}// end multiStop ()

	/**
	 * Default accessor for internal variable
	 * @return TwidorMain my tutor
	 */
	public TwidorMain myTutor () {
		return theTutor;
	}// end myTutor ()

	/**
	 * Default accessor for internal variable
	 * @return TwidorConstants the Defaults
	 */
	public TwidorConstants getDefaults () {
		return myDefaults;
	}// end getDefaults ()

	/**
	 * accessor for internal variable
	 * @return SerialTwiddler
	 */
	public SerialInterface getTwiddler () {
		return myTwiddler;
	}// end getTwiddler ()

// ActionListener Dependencies

	/**
	 * Invoked when an action is performed.
	 * @param e the Action to be interpreted
	 */
	public void actionPerformed (ActionEvent e) {
		if (!enabled) return;
	
		if (getDefaults().bDEBUG)
			System.out.println("EventHandler: Action Performed.");
		if (myHighlightTimer != null) {
			/* nasty, so don't ask */
			if (e.getSource().toString().equalsIgnoreCase(myHighlightTimer.toString())) {
				if (getDefaults().bDEBUG)
					System.out.println("EventHandler: Throwing highlight");
				myTutor().highlight();
				return;
			}
			if (e.getActionCommand() != null) {
				if (e.getActionCommand().equals(getDefaults().START_TEXT))
					myHighlightTimer.start();
			}
		}
		if (myMultiTapTimer != null) {
			if (e.getSource().toString().equalsIgnoreCase(myMultiTapTimer.toString())) {
				/* send a forced-forward */
				if (getDefaults().bDEBUG)
					System.out.println("EventHandler: Throwing next key");
				myTutor().multitapTyped();
				return;
			}
		}
		if (e.getActionCommand() != null) {
			myTutor().menuOption(e.getActionCommand());
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("EventHandler: Rogue Timer Event.");
		}
	}// end actionPerformed

// WindowListener Dependencies

	/**
	 * Invoked when the window is set active.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowActivated (WindowEvent e) {
	}// end windowActivated

	/**
	 * Invoked when dispose is called on a window.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowClosed (WindowEvent e) {
		myTutor().twidorQuit();
	}// end windowClosed

	/**
	 * Invoked when the window is closed from the menu.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowClosing (WindowEvent e) {
		myTutor().twidorQuit();
	}// end windowClosing

	/**
	 * Invoked when a window is unselected.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowDeactivated (WindowEvent e) {
	}// end windowDeactivated

	/**
	 * Invoked when a window is un-minimized.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowDeiconified (WindowEvent e) {
	}// end windowDeiconified

	/**
	 * Invoked when a window is minimized.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowIconified (WindowEvent e) {
	}// end windowIconified

	/**
	 * Invoked when a window is first made visible.
	 * @param e the WindowEvent to be interpreted
	 */
	public void windowOpened (WindowEvent e) {
	}// end windowOpened

// KeyListener Dependencies

	/**
	 * Invoked when a key has been pressed.
	 * @param e the KeyEvent to be interpreted
	 */
	public void keyPressed (KeyEvent e) {
	}// end keyPressed

	/**
	 * Invoked when a key has been released.d	 * @param e the KeyEvent to be interpreted
	 */
	public void keyReleased (KeyEvent e) {
	}// end keyReleased

	/**
	 * Invoked when a key has been typed.
	 * @param e the KeyEvent to be interpreted
	 */
	public void keyTyped (KeyEvent e) {
		if (!enabled) return;
		if (getDefaults().ACCEPT_INPUT) {
			if (getDefaults().bDEBUG)
				System.out.println("EventHandler: Key Pressed " + e.getKeyChar());
			myTutor().keyboardTyped(String.valueOf(e.getKeyChar()),
					System.currentTimeMillis());
			if (myHighlightTimer != null) {
				myHighlightTimer.restart();
			}
			if (myMultiTapTimer != null) {
				myMultiTapTimer.restart();
			}
		}
	}// end keyTyped

	/**
	 * invoked when a key has been typed via the serial port
	 * @param e the KeyElement the key that was pressed
	 */
	public void keyState (KeyElement buttons) {
		if (!enabled) return;
		if (getDefaults().ACCEPT_INPUT) {
			myTutor().serialState(buttons);
		}
	}// end (KeyElement)

	/**
	 * invoked when a key is to be generated
	 * @param e the KeyElement the key to be generated
	 */
	public void keyEvent (KeyElement buttons) {
		if (!enabled) return;
		if (getDefaults().ACCEPT_INPUT) {
			myTutor().serialTyped(buttons);
			if (myHighlightTimer != null) {
				myHighlightTimer.restart();
			}
			if (myMultiTapTimer != null) {
				myMultiTapTimer.restart();
			}
		}
	}// end keyEvent (KeyElement)

// ItemListener Dependencies
	
	/**
	 * Invoked when an item has been selected or deselected
	 * @param e the ItemEvent to be interpreted
	 */
	public void itemStateChanged (ItemEvent e) {
		if (!enabled) return;
		if (getDefaults().bDEBUG)
			System.out.println("EventHandler: Menu Item changed state");
		if (e.getItem() instanceof JCheckBoxMenuItem) {
			myTutor().booleanOption(((JCheckBoxMenuItem)e.getItem()).getText(),
					((JCheckBoxMenuItem)e.getItem()).getState());
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("EventHandler: Menu " + e.paramString());
		}
	}// end itemStateChanged

	public boolean isEnabled() {
		return enabled;
	}


}// end class EventHandler
