/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.io.*;
import java.util.*;
/**
 * CCG: Twidor- The Twiddler Tutor!
 * <pre>
 * KeyMap.java, a class for loading and storing of a KeyMap from the
 * 'standard' Twiddler1 format.
 *
 * Revisions:
 * 		0.5	17 July 2003
 * 		Completed Tutor
 * 		0.2	06 June 2003
 * 		Need to implement keymap file parsing.
 *		0.1	23 May 2003
 * 		Created class KeyMap
 * </pre>
 * @author <a href="mailto:visyz@cc.gatech.edu">Gibby Fusia</a>
 * @version Version 0.5; 17 July 2003
 */
public class KeyMap {

	/**
	 * the keymap, as we see it. a Vector of KeyElements (buttons + map)
	 */
	private Vector keylist;
	private Map keymap;
	/**
	 * the 'special' buttons we need to know about
	 */
	private Vector macrolist;
	/**
	 * the defaults we refer to, from time to time
	 */
	private TwidorConstants myDefaults;

	/**
	 * default constructor
	 */
	public KeyMap () {
		keylist = new Vector();
		macrolist = new Vector();
		keymap = new HashMap();
	}// end KeyMap

	/**
	 * default constructor
	 * @param twidorDefaults the defaults that we should use
	 */
	public KeyMap (TwidorConstants twidorDefaults) {
		this();
		setDefaults(twidorDefaults);
		readKeyMap(getDefaults().DEFAULT_KEYMAP);
	}// end KeyMap (String)

	/**
	 * Sets the defaults that this element should use
	 * @param twidorDefaults the constants to be used
	 */
	private void setDefaults (TwidorConstants twidorDefaults) {
		myDefaults = twidorDefaults;
	}// end setDefaults (TwidorConstants)

	/**
	 * Fetches the defaults
	 * @return TwidorConstants the defaults
	 */
	private TwidorConstants getDefaults () {
		return myDefaults;
	}// end getDefaults ()

	/**
	 * Gets the Keys matching the requested letter
	 * @param key the letter to match
	 * @return KeyElement the Key that matches the desired String
	 */
	public KeyElement getKey (String key) {
		KeyElement toReturn = null;
		
		if (key.length() > 1) {
			toReturn = (KeyElement) keymap.get(key);
		} else if (key.length() > 0) {
			int code = (int) key.charAt(0);
			toReturn = (KeyElement) keymap.get(new Integer(code));
		}

		/*
		int i = getKeylist().indexOf(new KeyElement(key));

		if (i >= 0) {
			toReturn = (KeyElement) getKeylist().elementAt(i);
		}
		*/
		return toReturn;
	}// end getKey (String)

	/**
	 * Gets the Key matching the KeyElement button vector set
	 * @param buttons - which buttons are pressed
	 * @return KeyElement the requested KeyElement
	 */
	public KeyElement getKey (int buttons) {
		KeyElement toReturn = null;
		//KeyElement test = new KeyElement(buttons);
		//int bits = test.getBits();

		for (int i=0; i<getKeylist().size(); i++) {
			KeyElement key = (KeyElement) getKeylist().elementAt(i);
			if (/*key.equals(test) || */key.getBits() == buttons) return key;
		}
		return null;

	}// end getKey (Vector)

	/**
	 * Accessor for the list of keys that can do macros
	 * @return Vector the list of macros (and key combinations)
	 */
	public Vector getMacros () {
		return macrolist;
	}// end getMacros ()

	/**
	 * Function for testing whether or not the KeyMap appears valid
	 * @return boolean the validity of the keymap
	 */
	public boolean appearsValid () {
		if (getKeylist().size() >= 26) {
			return true;
		}
		return false;
	}// end appearsValid ()

	/**
	 * gets the internal keylist
	 * @return Vector the known keymap
	 */
	private Vector getKeylist () {
		return keylist;
	}// end getKeylist ()

	/**
	 * Adds the KeyElement to the KeyMap
	 * @param key the desired key to add
	 * @return boolean the success or failure of the add
	 */
	public boolean addKey (KeyElement key) {
		if (keylist.indexOf(key) >= 0) {
			//return false;
			keylist.removeElementAt(keylist.indexOf(key));
		}

		keylist.addElement(key);

		if (key.getLetter() != null) {
			keymap.put(key.getLetter(), key);
		} else {
			keymap.put(new Integer(key.getNumber()), key);
			if (TwidorConstants.bDEBUG) System.out.println("KeyMap: Adding key: " + key.getNumber());
		}
		

		return true;
	}// end addKey (KeyElement)

	/**
	 * sets the new keymap
	 * @param source the source for the new keymap
	 */
	public void readKeyMap (String source) {
		InputStream iStream;
		InputStreamReader isReader;
		BufferedReader brBuffer;
		boolean finished = false;
		String temp;
		int test;

		if (getDefaults().bDEBUG)
			System.out.println("KeyMap: reading file " + source);
		
		try {
			/* FIXME: support reading from both a jar and a fs file */
			iStream = this.getClass().getResourceAsStream(source);
			isReader = new InputStreamReader(iStream);
			brBuffer = new BufferedReader(isReader);
			while (brBuffer.ready()) {
				temp = brBuffer.readLine();
				temp = remove_comment(temp, "#");
				temp = remove_comment(temp, ";");
				temp = remove_comment(temp, "!");
				temp = remove_comment(temp, "//");
				/* Comments removed */
				test = temp.indexOf('=');
				if (test > -1) {
					parseLine(temp);
				}
			}
			brBuffer.close();
			isReader.close();
			iStream.close();
		} catch (NullPointerException e) {
			System.out.println("Keymap File not found: " + source);
		} catch (IOException e) {
			if (getDefaults().bDEBUG)
				System.out.println("KeyMap: IO Error while reading " + source);
		} catch (Exception e) {
			if (getDefaults().bDEBUG)
				System.out.println("KeyMap: General Error: " + source);
		}

		if (getDefaults().bDEBUG)
			System.out.println("KeyMap: " + getKeylist().size() + " key bindings, " +
					getMacros().size() + " macros");
	}// end setKeyMap (String)

	/**
	 * function to remove comments
	 * @param toParse the line to parse
	 * @param comment the comment to look for and remove everything to the right of
	 * @return String the string sans comment
	 */
	private String remove_comment (String toParse, String comment) {
		int test;
		if ((test = toParse.indexOf(comment)) != -1) {
			if (getDefaults().bDEBUG)
				System.out.println("Removing Comment (" + comment + ")");
			return toParse.substring(0,test);
		}
		return toParse;
	}// end remove_comment (String, String)

	/**
	 * function to quickly convert letters to their appropriate button
	 * @param toCompare the character to look at
	 * @return int the value of this character
	 */
	private int finger_button (char toCompare) {
		switch (toCompare) {
			case 'L':	case 'l':
				return getDefaults().B_LEFT;
			case 'M':	case 'm':
				return getDefaults().B_MIDDLE;
			case 'R':	case 'r':
				return getDefaults().B_RIGHT;
			default:
				return -1;
		}
	}// end finger_button (char)

	/**
	 * parses a keymap line and attempts to make a button/letter
	 * association out of it.
	 * @param line the string to parse
	 */
	private void parseLine(String line) {
		String token;
		int pos1, pos2;
		KeyElement newKey = new KeyElement();
		StringTokenizer mainTokens = new StringTokenizer(line, "=");

		if (mainTokens.countTokens() != 2) {
			if (getDefaults().bDEBUG) {
				System.out.println("Not a valid entry");
			}
			return;
		}
		try {
			StringTokenizer LeftSide = new StringTokenizer(mainTokens.nextToken());
			StringTokenizer RightSide = new StringTokenizer(mainTokens.nextToken(), ",");

			if (LeftSide.countTokens() != 2) {
				if (getDefaults().bDEBUG)
					System.out.println("Malformed chord");
				return;
			}

			/* Thumb Modifiers */
			token = LeftSide.nextToken().trim();
			pos2 = 0;
			do {
				String temp;
				pos1 = token.indexOf('+', pos2);
				if (pos1 != -1) {
					temp = token.substring(pos2, pos1);
				}
				else {
					temp = token.substring(pos2);
				}
				/* Match up the thumb buttons */
				if (temp.equalsIgnoreCase("0")) {
				}
				else if (temp.equalsIgnoreCase("NUM")) {
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_NUM, true);
				}
				else if (temp.equalsIgnoreCase("ALT")) {
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_ALT, true);
				}
				else if (temp.equalsIgnoreCase("CTRL")) {
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_CTRL, true);
				}
				else if (temp.equalsIgnoreCase("SHIFT")) {
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_SHIFT, true);
				}
				else if (temp.equalsIgnoreCase("FUNC")) {
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_NUM, true);
					newKey.setButton(getDefaults().THUMB_OFFSET +
							getDefaults().B_ALT, true);
				}
				else {
		//			if (getDefaults().bDEBUG) System.out.println("Unknown token: " + temp);
				}
				pos2 = pos1 + 1;
			} while (pos1 != -1);

			/* Finger Modifiers */
			token = LeftSide.nextToken().trim();
			for (int finger = 0; finger < 4; finger++) {
				int ftemp = finger_button(token.charAt(finger));
				if (ftemp != -1) {
					newKey.setButton(finger * getDefaults().FINGER_OFFSET + ftemp, true);
				}
			}

			/* And what the chord produces */
			//while (RightSide.hasMoreTokens()) {
			token = RightSide.nextToken().trim();
			//if (token.matches("\"\\w*\"")) {
			if (token.startsWith("\"") && token.endsWith("\"")) {
				/* Looks like a macro */
				newKey.setLetter(token.replaceAll("\"", ""));
				/* add it to the macro list */
				getMacros().add(newKey);
			}
			else {
				/* Assume it's a number */
				int ctemp = Integer.parseInt(token.trim());
				if (ctemp != -1) {
					newKey.setNumber(ctemp);
				}
			}

			if (newKey.displayLetter() != null) {
				addKey(newKey);
			}
		}
		catch (Exception e) {
			if (getDefaults().bDEBUG) System.out.println("KeyMap: " + e.toString());
		}

	}// end parseLine (String)

}// end class KeyMap
