/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.io.*;
import java.util.*;
import javax.comm.*;
/**
 * CCG: Twidor- The Twiddler Tutor!
 * <pre>
 * SerialTwiddler.java, the interface to the serial version of the Twiddler.
 *
 * Revisions:
 *                      0.1     16 August 2003
 * </pre>
 * @author <a href="mailto:visyz@cc.gatech.edu">Gibby Fusia</a>
 * @version Version 0.1; 16 August 2003
 */
public class SerialTwiddler implements SerialPortEventListener, SerialInterface {

	/**
	 * internal variables
	 */
	private TwidorConstants myDefaults;
	private EventHandler myEvents;
	private static InputStream myInput;
	private static SerialPort myPort;
	private static KeyElement myState;
	private boolean clearable;

	/**
	 * default constructor
	 */
	private SerialTwiddler () {
	}// end SerialTwiddler

	/**
	 * default constructor
	 * @param TwidorConstants the settings to use
	 */

	public SerialTwiddler (TwidorConstants twidorDefaults, EventHandler myEH) {
		setDefaults(twidorDefaults);
		setEventHandler(myEH);
		setState(new KeyElement(getDefaults().KEY_NOT_A_KEY));
		setClearable(false);
		init();
	}// end SerialTwiddler (TwidorConstants)

	/**
	 * internal modifier
	 * @param TwidorConstants the constants to use
	 */
	private void setDefaults (TwidorConstants twidorDefaults) {
		myDefaults = twidorDefaults;
	}// end setDefaults (TwidorConstants)

	/**
	 * internal modifier
	 * @param EventHandler the eventHandler to use
	 */
	private void setEventHandler (EventHandler newEH) {
		myEvents = newEH;
	}// end setEventHandler (EventHandler)

	/**
	 * internal modifier
	 * @param KeyElement the new state to set
	 */
	private void setState (KeyElement newKey) {
		myState = newKey;
	}// end setState (KeyElement)

	/**
	 * internal modifier
	 * @param boolean the new state to set to the key
	 */
	private void setClearable (boolean status) {
		clearable = status;
	}// end setClearable (boolean)

	/**
	 * internal accessor
	 * @return TwidorConstants the constants to use
	 */
	private TwidorConstants getDefaults () {
		return myDefaults;
	}// end getDefaults ()

	/**
	 * internal accessor
	 * @return EventHandler the EventHandler
	 */
	private EventHandler getEventHandler () {
		return myEvents;
	}// end getEventHandler ()

	/**
	 * internal accessor
	 * @return KeyElement the internal state
	 */
	private KeyElement getState () {
		return myState;
	}// end getState ()

	/**
	 * internal accessor
	 * @return boolean the state of the key
	 */
	private boolean getClearable () {
		return clearable;
	}// end getClearable ()

	/**
	 * function for opening the serial port
	 */
	public void init () {
		CommPortIdentifier tempID;

		try {
			/* See if the port exists */
			tempID = CommPortIdentifier.getPortIdentifier(getDefaults().SERIAL_DEVICE);
			/* Try and claim it */
			myPort = (SerialPort) tempID.open("Twidor2", getDefaults().SERIAL_TIMEOUT);
			/* Configure it */
			myPort.setSerialPortParams(getDefaults().SERIAL_BAUD_RATE, SerialPort.DATABITS_8,
				SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
			/* necessary to force the twiddler to send data */
			myPort.setDTR(false);
			myPort.notifyOnDataAvailable(true);
			/* Grab the input stream */
			myInput = myPort.getInputStream();
			/* Add the eventhandler */
			myPort.addEventListener(this);
		} catch (NoSuchPortException e) { /* Port doesn't exist */
			System.out.println("Unable to find port " + getDefaults().SERIAL_DEVICE);
			return;
		} catch (UnsupportedCommOperationException e) { /* We can't work with you. Sorry! */
			System.out.println("Unable to configure serial port hardware");
			return;
		} catch (PortInUseException e) { /* Somebody else is using the port */
			System.out.println("Serial port in use by somebody not us");
			return;
		} catch (IOException e) { /* Can't get the InputStream from the port */
			System.out.println("Unable to get Input from port");
			return;
		} catch (Exception e) { /* */
			System.out.println("Generic error opening the Serial Port");
			return;
		}

		try { /* clear the buffer */
			myInput.skip(myInput.available());
		} catch (IOException e) {
			System.out.println("Error with the Input Stream");
			return;
		}
		if (getDefaults().bDEBUG)
			System.out.println("SerialTwiddler: Successfully opened " +
				getDefaults().SERIAL_DEVICE);
	}// end init ()

	/**
	 * function for closing the serial port
	 */
	public void close () {
		if (myInput != null) {
			try {
				myInput.close();
			} catch (IOException e) {
				if (getDefaults().bDEBUG)
					System.out.println("SerialTwiddler: Error closing Port");
			}
		}
		if (myPort != null) {
			if (getDefaults().bDEBUG)
				System.out.println("SerialTwiddler: Closing port " +
					getDefaults().SERIAL_DEVICE);
			myPort.close();
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("SerialTwiddler: Port was never opened");
		}
	}// end close ()

	/**
	 * function for reading data from the serial port
	 */
	public void readInput () {
		byte[] buffer = new byte[5];
		try {
			while (myInput.available() > 5) {
				myInput.read(buffer, 0, 1);
				for (int i = 0; i < 5; i++) {
					if (i == 0) {
						while ((buffer[i] & 0x80) != 0) {
							myInput.read(buffer, i, 1);
						}
					}
					else {
						myInput.read(buffer, i, 1);
						if ((buffer[i] & 0x80) == 0) {
							buffer[0] = buffer[i];
							i = 1;
						}
					}
				}
				//System.out.println((int)(buffer[1] & 0x7f) * 128 + (int)(buffer[0]));
				checkState((int)(buffer[1] & 0x7f) * 128 + (int)(buffer[0]));
			}
		} catch (Exception e) {
			if (getDefaults().bDEBUG)
				System.out.println("SerialTwiddler: Error " + e);
			e.printStackTrace();
		}
	}// end readInput ()

	/**
	 * determine if the serial port has done anything new.
	 */
	public void checkState (int result) {
		boolean changed = false;
		KeyElement newStatus = new KeyElement(getDefaults().KEY_NOT_A_KEY);

		if (result >= 8192) {// mouse
			result -= 8192;
		}
		if (result >= 4096) {// alt
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_ALT, true);
			result -= 4092;
		}
		if (result >= 2048) {// ctrl
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_CTRL, true);
			result -= 2048;
		}
		if (result >= 1024) {// func
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_NUM, true);
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_ALT, true);
			result -= 1024;
		}
		if (result >= 512) {// num
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_NUM, true);
			result -= 512;
		}
		if (result >= 256) {// shift
			newStatus.setButton(getDefaults().THUMB_OFFSET + getDefaults().B_SHIFT, true);
			result -= 256;
		}
		if (result >= 192) {// pinky R
			newStatus.setButton(getDefaults().PINKY_OFFSET + getDefaults().B_RIGHT, true);
			result -= 192;
		}
		else if (result >= 128) {// pinky M
			newStatus.setButton(getDefaults().PINKY_OFFSET + getDefaults().B_MIDDLE, true);
			result -= 128;
		}
		else if (result >= 64) {// pinky L
			newStatus.setButton(getDefaults().PINKY_OFFSET + getDefaults().B_LEFT, true);
			result -= 64;
		}
		if (result >= 48) {// ring R
			newStatus.setButton(getDefaults().RING_OFFSET + getDefaults().B_RIGHT, true);
			result -= 48;
		}
		else if (result >= 32) {// ring M
			newStatus.setButton(getDefaults().RING_OFFSET + getDefaults().B_MIDDLE, true);
			result -= 32;
		}
		else if (result >= 16) {// ring L
			newStatus.setButton(getDefaults().RING_OFFSET + getDefaults().B_LEFT, true);
			result -= 16;
		}
		if (result >= 12) {// middle R
			newStatus.setButton(getDefaults().MIDDLE_OFFSET + getDefaults().B_RIGHT, true);
			result -= 12;
		}
		else if (result >= 8) {// middle M
			newStatus.setButton(getDefaults().MIDDLE_OFFSET + getDefaults().B_MIDDLE, true);
			result -= 8;
		}
		else if (result >= 4) {// middle L
			newStatus.setButton(getDefaults().MIDDLE_OFFSET + getDefaults().B_LEFT, true);
			result -= 4;
		}
		if (result >= 3) {// index R
			newStatus.setButton(getDefaults().INDEX_OFFSET + getDefaults().B_RIGHT, true);
			result -= 3;
		}
		else if (result >= 2) {// index M
			newStatus.setButton(getDefaults().INDEX_OFFSET + getDefaults().B_MIDDLE, true);
			result -= 2;
		}
		else if (result >= 1) {// index L
			newStatus.setButton(getDefaults().INDEX_OFFSET + getDefaults().B_LEFT, true);
			result -= 1;
		}

		if (result != 0 && getDefaults().bDEBUG)
			System.out.println("SerialTwiddler: Missed a button option? " + result);

		boolean buttonPressed = false;
		boolean buttonReleased = false;
		KeyElement current = getState();
		for (int i = 0; i < newStatus.NUM_BUTTONS; i++) {
			if (newStatus.getButton(i) != current.getButton(i)) {
				if (newStatus.getButton(i)) { // button press
					buttonPressed = true;
				}
				else { // button release
					buttonReleased = true;
				}
			}
		}
		/* Strictly speaking, this thing should be timestamped way
		 * earlier. I don't think java is liking that though.
		 */
		if (buttonPressed || buttonReleased) {
			newStatus.setTime(System.currentTimeMillis());
		}
		/* Handle the state change */
		if (getClearable()) {
			if (buttonPressed) {
				setClearable(false);
			}
			if (buttonPressed || buttonReleased) {
				/* generate a fake key event */
				getEventHandler().keyState(newStatus);
				setState(newStatus);
			}
		}
		else {
			if (buttonReleased) {
				/* generate a real key event */
				current.setTime(newStatus.getTime());
				getEventHandler().keyEvent(current);
				getEventHandler().keyState(newStatus);
				setState(newStatus);
				setClearable(true);
			}
			else if (buttonPressed) {
				/* generate a fake key event */
				getEventHandler().keyState(newStatus);
				setState(newStatus);
				return;
			}
		}
		/* wash, rinse, repeat */
	}// end checkState ()


	// SerialPortEventListener Dependencies
	/**
	 * invoked whenever the serial port receives data
	 * @param e the SerialPortEvent the event that occurred
	 */
	public void serialEvent (SerialPortEvent e) {
		if (!getEventHandler().isEnabled()) return;
		if (e.getEventType() == SerialPortEvent.DATA_AVAILABLE) {
			getEventHandler().getTwiddler().readInput();
		}
	}// end serialEvent (SerialPortEvent)

}// end class SerialTwiddler ()
