/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.awt.*;
import javax.swing.*;
import java.util.Vector;
/**
 * CCG: Twidor- The Twiddler Tutor!
 * <pre>
 * TwiddlerPanel.java, the class that defines how the Twiddler face looks.
 *
 * Revisions:
 *		0.5	17 July 2003
 *		Completed Tutor
 * 		0.2	06 June 2003
 * 			Should be nearly ready.
 * 		0.1	29 May 2003
 * 			Created class TwiddlerPanel
 * </pre>
 * @author <a href="mailto:visyz@cc.gatech.edu">Gibby Fusia</a>
 * @version Version 0.5; 17 July 2003
 */
public class TwiddlerPanel extends JPanel {
	/**
	 * internal variables
	 */
	private KeyMap myKeyMap;
	private boolean thumbOrientation;
	private boolean fingerOrientation;
	private boolean showThumbMap;
	private boolean showFingerMap;
	private boolean showTwiddler;
	private Vector panelList;
	private TwidorConstants myDefaults;

	/**
	 * default constructor
	 */
	public TwiddlerPanel (TwidorConstants twidorDefaults, KeyMap newMap, boolean thumb, boolean finger) {
		setDefaults(twidorDefaults);

		if (getDefaults().bDEBUG)
			System.out.println("TwiddlerPanel: creating panel");

		setKeyMap(newMap);
		setBackground(getDefaults().TWIDDLER_BACKGROUND);
		setBorder(BorderFactory.createLineBorder(Color.BLACK));

		initPanels();
		/* general doodlings with the JPanel */
		setLayout(new GridLayout(5, 1));
		reOrient();

		if (getDefaults().bDEBUG)
			System.out.println("TwiddlerPanel: panel created");
	}// end TwiddlerPanel ()

	/**
	 * accessor for the defaults
	 * @return TwidorConstants the defaults
	 */
	private TwidorConstants getDefaults () {
		return myDefaults;
	}// end getDefaults ()

	/**
	 * accessor for the current keymap
	 * @return KeyMap the current keymap
	 */
	public KeyMap getKeyMap () {
		return myKeyMap;
	}// end getKeyMap ()

	/**
	 * accessor for the JPanels we contain
	 * @return Vector the vector of panels we contain
	 */
	public Vector getPanels () {
		return panelList;
	}// end getPanels ()

	/**
	 * modifier for the defaults
	 * @param TwidorConstants the new defaults
	 */
	private void setDefaults (TwidorConstants twidorDefaults) {
		myDefaults = twidorDefaults;
	}// end setDefaults (TwidorConstants)

	/**
	 * modifier for the current keymap
	 * @param KeyMap the desired KeyMap
	 */
	public void setKeyMap (KeyMap newMap) {
		myKeyMap = newMap;
	}// end setKeyMap (KeyMap)

	/**
	 * initializer for the panel list
	 */
	private void initPanels () {
		panelList = new Vector();
	}// end initPanels ()

	/**
	 * highlights the keys for a given character
	 * @param String the character to highlight
	 */
	public void highlight (String key) {
		KeyElement keyPress = getKeyMap().getKey(key);
		if (keyPress != null) {
			highlight(keyPress);
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("TwiddlerPanel: Can't find \"" + key + "\" to highlight");
		}
	}// end highlight (String)

	/**
	 * highlights the keys for the given KeyElement
	 * @param KeyElement the key to highlight
	 */
	public void highlight (KeyElement key) {
		if (key == null) {
			if (getDefaults().bDEBUG)
				System.out.println("TwiddlerPanel: Can't highlight. (null)");
			return;
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("TwiddlerPanel: Highlighting \"" +
						key.displayLetter() + "\"");
		}
		try {
			for (int finger = 0; finger < 4; finger++) {
				for (int button = 0; button < 3; button++) {
					if (key.getButton(finger * getDefaults().FINGER_OFFSET + button)) {
						((TwiddlerSubPanel)getPanels().elementAt(finger +
								1)).highlight(button,
								getDefaults().BUTTON_HIGHLIGHT);
					}
				}
			}
			for (int thumb = 0; thumb < 4; thumb++) {
				if (key.getButton(thumb + getDefaults().THUMB_OFFSET)) {
					((TwiddlerSubPanel)getPanels().elementAt(0)).highlight(thumb,
							getDefaults().BUTTON_HIGHLIGHT);
				}
			}
		}
		catch (ArrayIndexOutOfBoundsException e) {
			if (getDefaults().bDEBUG)
				System.out.println("TwiddlerPanel: array oob");
		}
	}// end highlight (KeyElement)

	/**
	 * clears all possibly highlighted keys
	 */
	public void clear () {
		for (int i = 0; i < getPanels().size(); i++) {
			((TwiddlerSubPanel)getPanels().elementAt(i)).clear();
		}
	}// end clear ()

	/**
	 * called whenever you want to do a redraw of the Twiddler display;
	 * i.e. anytime you modify the Finger/Thumb Orientation or KeyMap.
	 */
	public void reOrient () {
		if (getDefaults().bDEBUG)
			System.out.println("TwiddlerPanel: rearranging twiddler layout");
		setVisible(false);
		removeAll();
		setMinimumSize(new Dimension(getDefaults().TWIDDLER_X, getDefaults().WINDOW_Y));
		getPanels().clear();
		getPanels().addElement(new ThumbPanel(getDefaults(), getKeyMap()));
		for (int i = 0; i < 4; i++) {
			TwiddlerSubPanel sub;
			//if (TwidorConstants.USE_MULTITAP) sub = new MultiTapPanel(getDefaults(), getKeyMap(), i);
			/*else*/ sub = new FingerPanel(getDefaults(), getKeyMap(), i);
			getPanels().addElement(sub);
		}
		for (int i = 0; i < getPanels().size(); i++) {
			add((JPanel)getPanels().elementAt(i));
		}
		if (getDefaults().TWIDDLER_SHOW) {
			setVisible(true);
		}
		if (getDefaults().bDEBUG)
			System.out.println("TwiddlerPanel: rearrangement complete");
	}// end reOrient ()

}// end class TwiddlerPanel
