/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
/**
 * CCG: Twidor- The Twiddler Tutor!
 * <pre>
 * Twidor.java, the GUI of the program. Get Lots of glue, and a couple rolls
 * of ducttape.
 *
 * Revisions:
 *		0.5	17 July 2003
 *		Completed Tutor
 * 		0.1	22 May 2003
 * 			Created class Twidor
 * </pre>
 * @author <a href="mailto:visyz@cc.gatech.edu">Gibby Fusia</a>
 * @version Version 0.5; 17 July 2003
 */
public class TwidorMain extends JFrame {

	/**
	 * internal variables
	 */
	private EventHandler myEventHandler;
	private TwidorMenu myMenuBar;
	private TwiddlerPanel myTwiddler;
	private TypingPanel myTypingPanel;
	private StatsPanel myStatsPanel;
	private KeyMap myKeyMap;
	private Sentence mySentences;
	private TwidorConstants myDefaults;
	private boolean showingMessage;
	private static KeyElement myLastKey;
	private static int myLastKeyCount;
	

	/**
	 * The Default Constructor
	 * @param twidorDefaults the defaults this program should use
	 */
	public TwidorMain (TwidorConstants twidorDefaults) {
		setDefaults(twidorDefaults);
		/* JFrame Settings */
		setSize(new Dimension(getDefaults().WINDOW_X, getDefaults().WINDOW_Y));
		setTitle(getDefaults().WINDOW_TITLE);
		setBackground(getDefaults().WINDOW_BACKGROUND);
		setResizable(getDefaults().WINDOW_RESIZABLE);
		
		setEventHandler();

		readDefaults();

		if (getDefaults().RUNNING_STUDY) {
			getStatsPanel().enableInput(false);
			String message = getDefaults().callbacks.getMessage();
			getTypingPanel().displayStatus(message);
			showingMessage = true;
		}

		getDefaults().ACCEPT_INPUT = true;
		getDefaults().SENTENCE_NUMBER = 0;


	}// end Twidor

	private void readDefaults() {
		/* Root Panel Settings */
		Container tempRoot = getContentPane();
		tempRoot.removeAll();
		tempRoot.setLayout(new BorderLayout());

		showingMessage = false;
		//setEventHandler();
		getEventHandler().reinit();

		setSentences();

		setKeyMap();

		if (getDefaults().MENUBAR_SHOW) {
			setTwidorMenu();
		}
		if (getDefaults().TWIDDLER_SHOW) {
			setTwiddlerPanel(getDefaults().TWIDDLER_MIRROR,
				getDefaults().TWIDDLER_MIRROR);
		}

		setTypingPanel();
		setStatsPanel();

		setLastKey(null);
		setLastKeyCount(0);

		tempRoot.add(getTwiddlerPanel(), BorderLayout.WEST);

		JPanel CenterPane = new JPanel();
		CenterPane.setLayout(new GridLayout(2, 1));
		CenterPane.add(getTypingPanel());
		CenterPane.add(getStatsPanel());
		CenterPane.setBackground(Color.WHITE);
		CenterPane.setVisible(true);

		tempRoot.add(CenterPane, BorderLayout.CENTER);
		/* Show it all */
		setVisible(true);
	}

	/**
	 * Sets the defaults for the system
	 */
	private void setDefaults (TwidorConstants twidorDefaults) {
		myDefaults = twidorDefaults;
	}// end setDefaults ()

	/**
	 * Sets the EventHandler for the System
	 */
	private void setEventHandler() {
		myEventHandler = new EventHandler(this, getDefaults());
		addWindowListener(myEventHandler);
		addKeyListener(myEventHandler);
	}// end setEventHandler ()

	/**
	 * Sets the Keymap for the system
	 */
	private void setKeyMap () {
		myKeyMap = new KeyMap(getDefaults());
		if (!getDefaults().USE_MULTITAP) {
			/* ignore the appearsValid flag */
			if (!myKeyMap.appearsValid()) {
				System.out.println("Unable to laod keymap, exiting. (" +
						getDefaults().DEFAULT_KEYMAP + ")");
				System.exit(-1);
			}
		}
	}// end setKeymap (String)

	/**
	 * Sets the JMenuBar for the options
	 */
	private void setTwidorMenu () {
		myMenuBar = new TwidorMenu(getDefaults(), this, getEventHandler());
		setJMenuBar(myMenuBar);
	}// end setTwidorMenu ()

	/**
	 * Sets the TwiddlerPanel for the system
	 * @param boolean the Thumb Orientation
	 * @param boolean the Finger Orientation
	 */
	private void setTwiddlerPanel (boolean thumb, boolean finger) {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Adding TwiddlerPanel");
		myTwiddler = new TwiddlerPanel(getDefaults(), getKeyMap(), thumb, finger);
	}// end setTwiddlerPanel (boolean, boolean)

	/**
	 * Sets the TypingPanel up
	 */
	private void setTypingPanel () {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Adding TypingPanel");
		myTypingPanel = new TypingPanel(getDefaults());
	}// end setTypingPanel ()

	/**
	 * Set up the StatsPanel
	 */
	private void setStatsPanel () {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Adding StatsPanel");
		myStatsPanel = new StatsPanel(getDefaults());
	}// end setStatsPanel ()

	/**
	 * Sets the Sentences the system will use to tutor
	 * @param String the source of the sentences
	 */
	private void setSentences () {
		mySentences = new Sentence(getDefaults());
		if (mySentences.getSize() <= 0) {
			System.out.println("Unable to load sentences, exiting (" +
					getDefaults().DEFAULT_PHRASES + ")");
			System.exit(-1);
		}
		getDefaults().sentenceList = mySentences;
	}// end setSentences (String)

	/**
	 * sets the last key typed, for use in multitap
	 * @param KeyElement the last key typed
	 */
	private void setLastKey (KeyElement last) {
		if (last == null) {
			myLastKey = null;
		}
		else {
			myLastKey = (KeyElement)last.clone();
		}
	}// end setLastKey (KeyElement)

	/**
	 * sets the last key count
	 * @param int the number of times it was pressed
	 */
	private void setLastKeyCount (int last) {
		myLastKeyCount = last;
	}// end setLastKeyCount (int)

	/**
	 * Accessor for the Defaults
	 * @return TwidorConstants the defaults
	 */
	private TwidorConstants getDefaults () {
		return myDefaults;
	}// end getDefaults ()

	/**
	 * Accessor for the EventHandler
	 * @return EventHandler for the system
	 */
	private EventHandler getEventHandler () {
		return myEventHandler;
	}// end getEventHandler ()

	/**
	 * Accessor for the KeyMap
	 * @return KeyMap the current keymap
	 */
	private KeyMap getKeyMap () {
		return myKeyMap;
	}// end getKeymap ()

	/**
	 * Accessor for the TwidorMenu
	 * @return TwidorMenu the system's menubar
	 */
	private TwidorMenu getTwidorMenu () {
		return myMenuBar;
	}// end getTwidorMenu ()

	/**
	 * Accessor for the TwiddlerPanel
	 * @return TwiddlerPanel the TwiddlerPanel
	 */
	private TwiddlerPanel getTwiddlerPanel () {
		return myTwiddler;
	}// end void getTwiddlerPanel ()

	/**
	 * Accessor for the TypingPanel
	 * @return TypingPanel the typing panel
	 */
	private TypingPanel getTypingPanel () {
		return myTypingPanel;
	}// end getTypingPanel ()

	/**
	 * Accessor for the StatsPanel
	 * @return StatsPanel the stats panel
	 */
	private StatsPanel getStatsPanel () {
		return myStatsPanel;
	}// end getStatsPanel ()

	/**
	 * Gets the current sentence
	 * @return String the current sentence
	 */
	public String getSentence () {
		int current = getDefaults().SENTENCE_NUMBER;
		if (current > -3 && current < mySentences.getSize()) {
			return mySentences.getSentence(current);
		}
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Sentence choice out of range");
		return null;
	}// end getSentence ()

	/**
	 * fetches the last key typed
	 * @return KeyElement the last key typed. May return NULL
	 */
	public KeyElement getLastKey () {
		return myLastKey;
	}// end getLastKey ()

	/**
	 * Fetches the number of times the last key was typed.
	 */
	public int getLastKeyCount () {
		return myLastKeyCount;
	}// end getLastKeyCount ()

	/**
	 * Accessor for the EventHandler to inform the Tutor a key has been
	 * pressed.
	 * Logging for a state change is handled at this level.
	 * @param char the key pressed
	 * @param long the system time when it was pressed
	 */
	public void keyboardTyped (String key, long time) {
		KeyElement typed = getKeyMap().getKey(key);

		if (typed == null && key != null) {
			typed = new KeyElement((int) key.charAt(0));
		} else if (typed == null) {
			if (getDefaults().bDEBUG) {
				System.out.println("TwidorMain: Input null (!?)");
			}
			return;
		}
		typed.setTime(time);
		charEvent(typed);
	}// end keyboardTyped (String, long)

	/**
	 * Function for effecting a keypress
	 * Logging for a state change is handled at this level.
	 * @param KeyElement the key that was pressed
	 */
	public void serialTyped (KeyElement key) {
		KeyElement theLetter = getKeyMap().getKey(key.getBits());
		if (getDefaults().bDEBUG) {
			System.out.println("TwidorMain: charTyped: " + theLetter);
		}
		if (theLetter != null) {
			if (getDefaults().USE_MULTITAP) {
				key.setNumber(theLetter.getNumber());
			} else {
				if (theLetter.getLetter() != null) {
					if (theLetter.getLetter().length() <= 1 || getDefaults().ALLOW_MCC) {
						key.setLetter(theLetter.getLetter());
					} else {
						key.setNumber(getDefaults().KEY_NOT_A_KEY);
					}
				}
				else key.setNumber(theLetter.getNumber());
			}
			charEvent(key);
		}
		else {
			/* wasn't really a key, report it like a state */
			getStatsPanel().charState(key);
		}
	}// end serialTyped (KeyElement)

	/**
	 * Function for effecting a keypress
	 * called because the multitap timer went off
	 * Logging for a state change is handled at this level.
	 */
	public void multitapTyped () {
		/* force an advance */
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Multi-tap Key Timeout");
		if (getLastKey() != null) {
			if (getLastKey().getNumber() == getDefaults().KEY_SPACE ||
					getLastKey().getNumber() == getDefaults().KEY_MENTER ||
					getLastKey().getNumber() == getDefaults().KEY_LEFT ||
					getLastKey().getNumber() == getDefaults().KEY_RIGHT) {
				if (getDefaults().bDEBUG)
					System.out.println("TwidorMain: Ignoring key timeout");
				return;
			}
		}
		setLastKey(null);
		setLastKeyCount(0);
		getStatsPanel().charState(new KeyElement(getDefaults().KEY_MULTITAP));
		getTypingPanel().charTyped(getDefaults().KEY_RIGHT);
	}// end multitapTyped ()

	/**
	 * Function for recording a keypress
	 * Logging for a state change is handled at this level.
	 * @param KeyElement the key that was pressed
	 */
	public void serialState (KeyElement key) {
		if (getDefaults().bDEBUG) System.out.println("TwidorMain: Status key: " + key.getNumber());
		getStatsPanel().charState(key);
	}// end serialState (KeyElement)

	/**
	 * function for making a typed character go to the screen, and stuff
	 * Logging for a full key is handled at this level
	 * @param KeyElement the key that was pressed
	 */
	public void charEvent (KeyElement key) {
		/* Register the keypress */
		if (getDefaults().USE_MULTITAP) {
			KeyElement tapKey = (KeyElement)key.clone();
			/* figure out if we need to fiddle with the key any */
			if (getLastKey() == null) {
				if (key.getNumber() == getDefaults().KEY_RIGHT) {
					getTypingPanel().charTyped(getDefaults().KEY_SPACE);
					getDefaults().CHARACTER_NUMBER--;
				}
				setLastKey(key);
				setLastKeyCount(0);
			}
			else {
				if (key.equals(getLastKey())) {
					if (key.getNumber() == getDefaults().KEY_SPACE ||
							key.getNumber() == getDefaults().KEY_MENTER ||
							key.getNumber() == getDefaults().KEY_LEFT) {
						/* 1-letter buttons */
						setLastKeyCount(0);
					}
					else if (key.getNumber() == getDefaults().KEY_RIGHT) {
						getTypingPanel().charTyped(getDefaults().KEY_SPACE);
						getDefaults().CHARACTER_NUMBER--;
					}
					else if (key.getNumber() == 112 || key.getNumber() == 119) {
						/* 4-letter buttons */
						setLastKeyCount((getLastKeyCount() + 1) % 4);
					}
					else { /* 3-letter buttons */
						setLastKeyCount((getLastKeyCount() + 1) % 3);
					}
				}
				else {
					if (getLastKey().getNumber() == getDefaults().KEY_LEFT ||
							getLastKey().getNumber() == getDefaults().KEY_RIGHT ||
							getLastKey().getNumber() == getDefaults().KEY_SPACE) {
						if (key.getNumber() == getDefaults().KEY_RIGHT) {
							getTypingPanel().charTyped(getDefaults().KEY_SPACE);
							getDefaults().CHARACTER_NUMBER--;
						}
					}
					else if (key.getNumber() != getDefaults().KEY_RIGHT &&
							getLastKey().getNumber() != getDefaults().KEY_MENTER) {
						/*if (getDefaults().bDEBUG) System.out.println("Are we here?: " +
										key.getNumber() + " Last key: " + getLastKey());*/
						getTypingPanel().charTyped(getDefaults().KEY_RIGHT);
					}
					setLastKey(key);
					setLastKeyCount(0);
				}
			}
			tapKey.setNumber(key.getNumber() + getLastKeyCount());
			getStatsPanel().charTyped(tapKey);
			getTypingPanel().charTyped(tapKey.getNumber());
		}
		else {
			getStatsPanel().charTyped(key);
			//getTypingPanel().charTyped(key.getNumber());
			if (key.getLetter() == null || key.getLetter().length() == 1) {
				getTypingPanel().charTyped(key.getNumber());
			} else {
				for (int i=0; i<key.getLetter().length(); i++) {
					getTypingPanel().charTyped((int) key.getLetter().charAt(i));
				}
			}
		}

		
		boolean overrideEnter = false;
		if (getDefaults().TIMING_DATA) {
			if (getDefaults().timingStarted == 0 && key.getNumber() > 0) {
				getDefaults().timingStarted = System.currentTimeMillis();
			}

			if (getDefaults().timingStarted != 1 && System.currentTimeMillis() - getDefaults().timingStarted > 1000 * 10) {
				overrideEnter = true;
			} else if (key.getNumber() == 96) {
				overrideEnter = true;
			}
		} else if (showingMessage) {
			if (getDefaults().USE_MULTITAP) overrideEnter = (key.getNumber() == getDefaults().KEY_LEFT);
			else overrideEnter = (key.getNumber() == 104);	// h key
		}
		

		/* Determine if we need to change sentences */
		if ((getTypingPanel().sentenceComplete() && !getDefaults().TIMING_DATA) || overrideEnter) {
			/*
			if (getDefaults().USE_MULTITAP) {
				myEventHandler.multiStop();
			}
			*/

			int result = getDefaults().CODE_NEXT_SENTENCE;
			if (getDefaults().RUNNING_STUDY) {
				if (!showingMessage) {
					result = getDefaults().callbacks.sentenceDone(getStatsPanel().currentStats());
				} else {
					showingMessage = false;
					result = getDefaults().callbacks.sentenceDone(null);
				}
			} else {
				//getDefaults().SENTENCE_NUMBER++;
				int num = (int) (Math.random() * getDefaults().sentenceList.getSize());
				getDefaults().SENTENCE_NUMBER = num;
			}

			if (result == getDefaults().CODE_NEXT_SENTENCE) {
				if (getDefaults().TIMING_DATA) {
					getTypingPanel().displayStatus(getDefaults().callbacks.getMessage());
				} else {
					getTypingPanel().displaySentence(getSentence());
				}
				getStatsPanel().nextSentence(getSentence());
				getStatsPanel().enableInput(true);
			} else if (result == getDefaults().CODE_NEXT_BLOCK) {
				getStatsPanel().reLabel(true);
				getStatsPanel().enableInput(false);
				String message = getDefaults().callbacks.getMessage();
				getTypingPanel().displayStatus(message);
				showingMessage = true;
			} else if (result == getDefaults().CODE_SWITCH_INPUTS) {
				//getEventHandler().cleanup();
				if (getDefaults().bDEBUG)
					System.out.println("Rereading defaults...");
				readDefaults();
				//getTypingPanel().displaySentence(getSentence());
				//getStatsPanel().nextSentence(getSentence());
				getStatsPanel().reLabel(true);
				getStatsPanel().enableInput(false);
				String message = getDefaults().callbacks.getMessage();
				getTypingPanel().displayStatus(message);
				showingMessage = true;
			} else if (result == getDefaults().CODE_END) {
				twidorQuit();
			}
		}
		/* check out our highlighting rules */
		if (getDefaults().HIGHLIGHT_HINT && getDefaults().HIGHLIGHT_DELAY > 0) {
			getTwiddlerPanel().clear();
			return;
		}
		if (getDefaults().HIGHLIGHT_POST) {
			getTwiddlerPanel().clear();
			getTwiddlerPanel().highlight(key);
		}
		else if (getDefaults().HIGHLIGHT_HINT) {
			highlight();
		}
	}// end charEvent (char, long)

	/**
	 * function for running the highlighting system
	 */
	public void highlight () {
		if (getDefaults().bDEBUG)
			System.out.println("Looking for a key to highlight");
		getTwiddlerPanel().clear();
		if (getDefaults().HIGHLIGHT_MACRO) {
			/* search for macros to highlight */
			for (int i = 0; i < getKeyMap().getMacros().size(); i++) {
				KeyElement temp = (KeyElement)getKeyMap().getMacros().elementAt(i);
				if (getSentence().regionMatches(getDefaults().CHARACTER_NUMBER,
							temp.getLetter(), 0,
							temp.getLetter().length())) {
					/* matches */
					getTwiddlerPanel().highlight(temp);
					return;
				}
			}
		}
		getTwiddlerPanel().highlight(getTypingPanel().getHighlight());
	}// end highlight ()

	/**
	 * function for effecting menubar based changes
	 * @param String the MenuItem that was selected
	 * @param boolean its new status
	 */
	public void booleanOption (String option, boolean status) {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Changing " + option + ":" + status);
		getTwidorMenu().itemSelected(option, status);
		if (option.equals(getDefaults().TWIDDLER_SHOW_TEXT)) {
			getDefaults().TWIDDLER_SHOW = status;
			getTwiddlerPanel().reOrient();
		}
		else if (option.equals(getDefaults().TWIDDLER_SHOW_LETTERS_TEXT)) {
			getDefaults().TWIDDLER_SHOW_LETTERS = status;
			getTwiddlerPanel().reOrient();
		}
		else if (option.equals(getDefaults().TWIDDLER_MIRROR_TEXT)) {
			getDefaults().TWIDDLER_MIRROR = status;
			getTwiddlerPanel().reOrient();
		}
		else if (option.equals(getDefaults().HIGHLIGHT_ERRORS_TEXT)) {
			getDefaults().HIGHLIGHT_ERRORS = status;
		}
		else if (option.equals(getDefaults().ALLOW_ERRORS_TEXT)) {
			getDefaults().ALLOW_ERRORS = status;
		}
		else if (option.equals(getDefaults().HIGHLIGHT_HINT_TEXT)) {
			getDefaults().HIGHLIGHT_HINT = status;
			getDefaults().HIGHLIGHT_POST = !status;
		}
		else if (option.equals(getDefaults().HIGHLIGHT_POST_TEXT)) {
			getDefaults().HIGHLIGHT_POST = status;
			getDefaults().HIGHLIGHT_HINT = !status;
		}
		else if (option.equals(getDefaults().STATS_SHOW_TEXT)) {
			getDefaults().STATS_SHOW = status;
			/* FIXME: refresh the existance of the stats panel */
		}
		else if (option.equals(getDefaults().LOG_ENABLE_TEXT)) {
			getDefaults().LOG_ENABLE = status;
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("TwidorMain: Unhandled option");
		}
	}// end booleanOption (String, boolean)

	/**
	 * function for effecting menubar based changes
	 * @param String the menu option that was selected
	 */
	public void menuOption (String option) {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Option " + option);
		getTwidorMenu().itemSelected(option, true);
		if (option.equals(getDefaults().START_TEXT)) {
			getDefaults().ACCEPT_INPUT = true;
			int num = (int) (Math.random() * getDefaults().sentenceList.getSize());
			getDefaults().SENTENCE_NUMBER = num;
			getTypingPanel().displaySentence(getSentence());
			getStatsPanel().nextSentence(getSentence());
		}
		else if (option.equals(getDefaults().STOP_TEXT)) {
			getDefaults().ACCEPT_INPUT = false;
			getTypingPanel().removeAll();
			getTypingPanel().repaint();
		}
		else if (option.equals(getDefaults().START_FROM_TEXT)) {
			// Quick quizbox asking what number
		}
		else if (option.equals(getDefaults().QUIT_TEXT)) {
			twidorQuit();
		}
		else {
			if (getDefaults().bDEBUG)
				System.out.println("TwidorMain: Unhandled option");
		}
	}// end menuOption (String)

	/**
	 * function for cleaning up and closing and stuff
	 */
	public void twidorQuit () {
		if (getDefaults().bDEBUG)
			System.out.println("TwidorMain: Exiting Twidor.");
		setVisible(false);
		// insert stats saving stuff here.
		getStatsPanel().save();
		getEventHandler().cleanup();
		dispose();
		System.exit(0);
	}// end twidorQuit ()

}// end class TwidorMain
