/* Twidor - The twiddler tutor (among other things).
 * Copyright (C) 2003 James Fusia
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.text.DecimalFormat;

public class TypingModel
{
	protected KeyMap keymap;
	public static final double KEY_WIDTH = 0.2;
	public static final double KEY_SPACING = 0.25;

	public static final double KEY_REPEAT_TIME = .090;
	public static final double KEY_HELD_TIME = .115;

	/*
	// Averages from users
	public static final double [] FITTS_SLOPES = new double[] {.027, .045, .070, .002};
	public static final double [] FITTS_INT = new double [] {.062, .026, -.008, .184};
	*/
	public static final double [] FITTS_SLOPES = new double[] {.027, .045, .070, .002};
	public static final double [] FITTS_INT = new double [] {.062, .026, -.008, .184};

	//	The time it takes to press a key when that finger previously wasn't pressing anything
	public static final double [] REACTION_SPEED = new double[] {0,0,0,0};	

	DecimalFormat fmt = new DecimalFormat("###.##");

	public TypingModel(KeyMap keymap) {
		this.keymap = keymap;
	}

	public double predictSentenceDelay(String sentence) {
		double ret = KEY_HELD_TIME;	//	Length of time the first character is held down
		for (int i=0; i<sentence.length() - 1; i++) {
			char ch1 = sentence.charAt(i);
			char ch2 = sentence.charAt(i + 1);
			ret += getDelay(ch1, ch2);
		}

		return ret;
	}


	//	Returns predicted delay in seconds that it will take to move
	//	from ch1 to ch2.  Includes the time taken to press ch2, but
	//	not ch1.
	public double getDelay(char ch1, char ch2) {
		KeyElement k1 = keymap.getKey(String.valueOf(ch1));
		KeyElement k2 = keymap.getKey(String.valueOf(ch2));

		int [] fingerPos1 = new int[4];
		int [] fingerPos2 = new int[4];
		double [] transitionDelays = new double[4];

		if (k1 == null) {
			System.out.println("Model.getDelay(): Keymap didn't include character: " + ch1);
			return Double.NaN;
		} else if (k2 == null) {
			System.out.println("Model.getDelay(): Keymap didn't include character: " + ch1);
			return Double.NaN;
		}

		for (int i=0; i<4; i++) {
			fingerPos1[i] = k1.getFingerPosition(i);
			fingerPos2[i] = k2.getFingerPosition(i);
		}

		for (int i=0; i<4; i++) {
			transitionDelays[i] = getSingleFingerDelay(i, fingerPos1[i], fingerPos2[i]);
		}

		double max = 0;
		for (int i=0; i<4; i++) {
			if (transitionDelays[i] > max)
				max = transitionDelays[i];
		}

		double total = max + KEY_HELD_TIME;

		//if (ch1 == 'a') System.out.println();
		//System.out.print("" + ch1 + ch2 + ": " + fmt.format(total * 1000) + " ");

		return total;

	}

	public double getSingleFingerDelay(int finger, int pos1, int pos2) {
		if (pos1 == pos2) {
			if (pos1 == TwidorConstants.B_NONE) {
				//	This finger is not involved in the motion, so return zero
				return 0;
			} else {
				//	Same key repeated
				return KEY_REPEAT_TIME;
			}
		} else if (pos1 != TwidorConstants.B_NONE && pos2 != TwidorConstants.B_NONE) {
			//	finger moved
			return FITTS_SLOPES[finger] * getID(pos1, pos2) + FITTS_INT[finger];

		} else {
			if (pos1 == TwidorConstants.B_NONE) {
				//	A key was released
				return 0;
			} else {
				//	A key was pressed
				return REACTION_SPEED[finger];
			}
		}
	}

	//	Get Fitt's law index of difficulty between two finger positions
	//	Assume that both pos1 and pos2 are either B_LEFT, B_MIDDLE, or B_RIGHT
	public double getID(int pos1, int pos2) {
		int distance = Math.abs(pos1 - pos2);
		double A;
		double W = KEY_WIDTH;

		if (distance == 1) {
			//	From the middle of one key to the middle of the next
			A = KEY_SPACING + KEY_WIDTH;
		} else if (distance == 2) {
			A = 2 * KEY_SPACING + 2 * KEY_WIDTH;
		} else {
			System.out.println("Model.getID(): bad distance: " + distance);
			return Double.NaN;
		}

		double ret = Math.log(2*A/W)/Math.log(2);
		//System.out.println("TypingModel.getID returning " + ret);

		return ret;
	}

}
